#ifndef __TRC_BOARD_1_H__
#define __TRC_BOARD_1_H__

/*!
\file TRCBoard1.h
\brief First attempt to extend API for qfix SoccerBoard.

Project:	TRC SoccerBoard API

Programmer:	Jacek "TeMPOraL" Zlydach	[temporal.pl@gmail.com]

Created: 06.06.2009
Last Revised:	14.06.2009

Implementation File: None

====  LICENCE  ====
This file is distributed under TRC Postcard Licence v0.5.
You can use this file for anything you want (including commercial products), as long as:
*	Your use is ethical.
*	You don't remove this comment from the source.

You are allowed (and encouraged) to change, modify and fix things in this code.
If you find this code useful I'd be happy if you sent me a postcard.
Just mail me for address.
====  LICENCE  ====
	
*/

/*
	...
*/

#include "qfixSoccerBoard.h"

//! \brief Main API class.
//!
//! Contains original SoccerBoard inside.
//! See example file: trc_roam1.cc for usage.
class TRCBoard1
{
protected:
	SoccerBoard board;	//!< Original SoccerBoard API object.

	//Callibration data for distance sensors:
	enum {IR_SENSOR_SAMPLES = 10};	//!< How many distance samples we've collected.
	float irSensorRanges[IR_SENSOR_SAMPLES];	//!< Ranges (in meters)
	int irSensorSettings[IR_SENSOR_SAMPLES];	//!< Threshold values
	
	//Random number generator stuff
	int seed;
	
	int advSeed;
	int advGen1;
	int advGen2;
	int advNewSeed;
	
public:

	//! \brief Default c-tor
	TRCBoard1()
	{
		//set up ir sensor data
		irSensorSettings[0] = 0; 	irSensorRanges[0] = 1.3f;
		irSensorSettings[1] = 12;	irSensorRanges[1] = 0.65f;
		irSensorSettings[2] = 15;	irSensorRanges[2] = 0.34f;
		irSensorSettings[3] = 25;	irSensorRanges[3] = 0.30f;
		irSensorSettings[4] = 30;	irSensorRanges[4] = 0.24f;
		irSensorSettings[5] = 40;	irSensorRanges[5] = 0.16f;
		irSensorSettings[6] = 50;	irSensorRanges[6] = 0.13f;
		irSensorSettings[7] = 75;	irSensorRanges[7] = 0.08f;
		irSensorSettings[8] = 100;	irSensorRanges[8] = 0.06f;
		irSensorSettings[9] = 150;	irSensorRanges[9] = 0.025f;
		
		//rand
		seed =  1337;
		
		//advrand init
		advSeed = 1337;
		advGen2 = 3791;
		advGen1 = advGen2/2;
	}
	
	//! \brief Function to get access to original qfix SoccerBoard API.
	//! \return Returns reference to original qfix SoccerBoard API object.
	SoccerBoard& GetMainBoard()
	{
		return board;
	}
	
	//General use functions
	
	//! \brief Activates program start-up sequence.
	//!
	//! A common and good way to start your program. This function activates
	//! one LED and then waits for user to press button 0. This way your program
	//! won't start before you disconnect programming unit from the board.
	void StartupSequence()
	{
		board.ledOn(0);
		board.waitForButton(0);
		board.ledOff(0);
	}
	
	//! \brief Simple random number generator.
	//!
	//! Credit to one matematician I met in Galeria Krakowska; I can't remember
	//! your name, sorry :(.
	//! Still, this generator seems to be broken.
	//! \return Returns a pseudo-random number between 0 and 65536.
	int BasicRand()
	{
		return seed = (seed*seed) % 65537;
	}
	
	//! \brief A better random number generator.
	//!
	//! Credit to Game Programming Gems series, it's from there ;).
	//! I haven't time to test it though.
	//! \return Retruns a pseudo-random number between 0 and 65536.
	int AdvRand()
	{
		advNewSeed = ((advGen1 * advSeed) + advGen2);
		advSeed = advNewSeed;
		return advNewSeed % 65537;
	}
	
	//Movement functions
	//! \brief Move robot forward at given speed.
	//! \param speed - speed in -255 to 255; 0 to stop, negative values cause
	//!					the robot to move backwards.
	//!
	//!	This function just encapsulates a common set of function calls that
	//! cause robot to move forward.
	void MoveForward(int speed)
	{
		board.motor(0, -speed);
		board.motor(1, speed);
		board.motor(2, 0);
	}
	
	//! \brief Move robot backward at given speed.
	//! \param speed - speed in -255 to 255; 0 to stop, negative values cause
	//!					the robot to move forward.
	//!
	//!	This function just encapsulates a common set of function calls that
	//! cause robot to move backward.
	void MoveBackward(int speed)
	{
		MoveForward(-speed);
	}
	
	//! \brief Rotate robot left in-place (with no forward/backward movement).
	//! \param speed - speed in -255 to 255; 0 to stop, negative values cause
	//!					the robot to rotate right.
	//!
	//!	This function just encapsulates a common set of function calls that
	//! cause robot to rotate left.
	void RotateLeft(int speed)
	{
		board.motor(0, speed);
		board.motor(1, speed);
		board.motor(2, speed);
	}
	
	//! \brief Rotate robot right in-place (with no forward/backward movement).
	//! \param speed - speed in -255 to 255; 0 to stop, negative values cause
	//!					the robot to rotate left.
	//!
	//!	This function just encapsulates a common set of function calls that
	//! cause robot to rotate right.
	void RotateRight(int speed)
	{
		RotateLeft(-speed);
	}
	
	//Sensoric functions
	//! \brief Gives a threshold value of distance sensor for given distance. If a distance sensor
	//!			returns you a value, and that value is greater (>) than computed threshold it means
	//!			that there is an object closer to the sensor than the range you used to compute threshold.
	//! \param distance - in meters, the distance for which to compute threshold value.
	//! IR Distance Sensors give you value between 0 (nothing detected) to 255 (something directly ahead
	//! of the sensor); the closer an object is to the sensor, the bigger the return value will be.
	//! Therefore to check if an object is closer than given distance you use > operator (greater-than) to
	//! compare sensor readings with computed threshold.
	//!
	//! Range-to-Threshold computation is based on a linear interpolation between a set of points I have
	//! measured experimentally. Accuracy is not perfect, but is good enough :).
	int GetIRThresholdForDistance(float distance)
	{
		for(int i = 1 ; i < IR_SENSOR_SAMPLES ; ++i)
		{
			if(distance > irSensorRanges[i])
			{
				float scale = (distance - irSensorRanges[i])/(irSensorRanges[i-1] - irSensorRanges[i]);
				return static_cast<int>(irSensorSettings[i]*(scale) + irSensorSettings[i-1]*(1.0f-scale));
			}
		}
		return 0;
	}
};


#endif //__TRC_BOARD_1_H__

