//Simple patrolling code.
//================================================================
//Author: Jacek "TeMPOraL" Zlydach.
//(<C) 2009 - Eternity, Temporal Research Centre
//Last Update: 14.06.2009
//================================================================
//The goal of this program is to make the robot capable of wandering around
//both on the arena filled with other robots and on the floor between people
//(with little emphasis on the second option :) ). The robot will use its
//sensors to avoid collisions while doing some random roaming around.
//Best used with a video camera attached to the board ;)
//================================================================
//If you have questions, problems or suggestions to report, feel free to write
//me an e-mail: temporal.pl@gmail.com
//
//More details about this code and other robotics stuff at:
//http://temporal.pr0.pl/devblog/
//================================================================


//The core of this program is a simple state machine and an infinite loop
//that calls currently selected state to update.

//ROBOT CONFIGURATION:
//This program assumes that you are using SoccerBoard with
//following port configuration:
//ANALOG 0 - buzzer
//ANALOG 4 - front-left distance sensor [configurable]
//ANALOG 6 - front-right distance sensor [configurable]
//ANALOG 7 - back distance sensor [configurable]

//Assumptions from TRCBoard1:
//MOTOR0 - front-left motor
//MOTOR1 - front-right motor
//MOTOR2 - back motor

#include "TRC/TRCBoard1.h"

//Constants that tell which distance sensor is binded to which analog port.
//You can reconfigure them here.
enum
{
	FL_DISTANCE = 4,	//front-left distance sensor
	FR_DISTANCE = 6,	//front-right distance sensor
	BACK_DISTANCE = 7	//back distance sensor
};

//if anything is detected closer than following values (in meters),
//a corresponding state will be activated
const float ReactionDist_Evade = 0.25f;	//distance for evade state
const float ReactionDist_RunLikeHell = 0.10f;	//distance for run-like-hell state

//A very simple state machine composed of pointer to state function. One can change states
//by altering this pointer.
typedef void(* ptrFun_t) (TRCBoard1& board);	//pointer type definition
ptrFun_t currentState = 0;	//current state

//Following thresholds are computed from distance constants. They also represent
//distance, but in distance sensor range (0...255).
int ReactionThreshold_Evade = 0;
int ReactionThreshold_RunLikeHell = 0;

//Variables added as a hot-fix for broken random number generator.
//Used only with SignalPresense and LookAround states, they delay
//state-change so that those two states won't be activated too often.
int Cooldown_SignalPresense = 3;
int Cooldown_LookAround = 10;

//Forward declarations of state functions
void State_WanderingAround(TRCBoard1& board);
void State_Evade(TRCBoard1& board);
void State_RunLikeHell(TRCBoard1& board);

void State_SignalPresense(TRCBoard1& board);
void State_LookAround(TRCBoard1& board);

//Checks if anything is seen closer than given distance.
//Returns true if something is closer than given threshold, false otherwise.
bool CheckAllSensorDistance(TRCBoard1& board, int threshold)
{
	return ( ( board.GetMainBoard().analog(FL_DISTANCE) > threshold ) ||
		( board.GetMainBoard().analog(FR_DISTANCE) > threshold) ||
		( board.GetMainBoard().analog(BACK_DISTANCE) > threshold ) );
}

//STATE
//Wandering Around
//Basic and initial state of the robot. As the name suggests, robot actually
//wanders around, moving forward and sometimes turning.
//State transition occur to Evade when something is closer than Evade distance, and
//on random to LookAround and SignalPresence states (those two are only amusers ;))
void State_WanderingAround(TRCBoard1& board)
{
	//Move for 2 seconds, checking collisions 10 times per second.
	for(int i = 0 ; i < 20 ; ++i)
	{
		if( CheckAllSensorDistance(board, ReactionThreshold_Evade) )
		{
			currentState = State_Evade;
			return;
		}
	
		board.MoveForward(100);
		board.GetMainBoard().msleep(100);
	}
	
	//Randomize rotation direction and turn for 0.5 seconds, checking
	//collisions every 100 msec.
	if( CheckAllSensorDistance(board, ReactionThreshold_Evade) )
	{
		currentState = State_Evade;
		return;
	}
	
	board.RotateLeft( ((board.BasicRand() % 2)*2 - 1) * 75);
	for(int i = 0 ; i < 5 ; ++i)
	{
		if( CheckAllSensorDistance(board, ReactionThreshold_Evade) )
		{
			currentState = State_Evade;
			return;
		}
		board.GetMainBoard().msleep(100);
	}
	
	//Switch to special state. Cooldown introduced,
	//because BasicRand() is broken.
	if( (board.BasicRand() % 1000) <= 2)
	{
		if(--Cooldown_SignalPresense == 0)
		{
			Cooldown_SignalPresense = 50;
			currentState = State_SignalPresense;
		}
	}
	else if( (board.BasicRand() % 1000) <= 2)
	{
		if(--Cooldown_LookAround == 0)
		{
			Cooldown_LookAround = 50;
			currentState = State_LookAround;
		}
	}
	
}

//STATE
//Evade
//State is activated when something is too close to the robot. The best direction will
//be chosen, and the robot will try to avoid potential collision.
//State transition may occur to Wandering Around, if there is nothing on sensors closer
//than Evade distance, and to Run-Like-Hell if something gets really to close.
void State_Evade(TRCBoard1& board)
{
	//First, check if something is too close, so we should panic and... run like hell :).
	if( CheckAllSensorDistance(board, ReactionThreshold_RunLikeHell ) )
	{
		currentState = State_RunLikeHell;
		return;
	}
	
	//Following conditions check which sensor detects an object that is too close,
	//and choose the best direction to turn while still moving forward.
	//Robot gets blind for at least 100msec.
	if( board.GetMainBoard().analog(FL_DISTANCE) > ReactionThreshold_Evade)
	{
		//Right-forward
		board.GetMainBoard().motor(0, -100);
		board.GetMainBoard().motor(1, -200);
		board.GetMainBoard().motor(2, -50);
		board.GetMainBoard().msleep(100);
		board.GetMainBoard().motor(2, 0);
	}
	else if(board.GetMainBoard().analog(FR_DISTANCE) > ReactionThreshold_Evade)
	{
		//Left-forward
		board.GetMainBoard().motor(0, -100);
		board.GetMainBoard().motor(1, -200);
		board.GetMainBoard().motor(2, +50);
		board.GetMainBoard().msleep(100);
		board.GetMainBoard().motor(2, 0);
	}
//  Processing of back sensor turned off due to some problems.
//	else if(board.GetMainBoard().analog(BACK_DISTANCE) > ReactionThreshold_Evade)
//	{
//		//steruj przed siebie
//		board.MoveForward(100);
//		board.GetMainBoard().msleep(100);
//	}

	//If there is nothing close enought for us to evade it, we return to the
	//Wandering Around state.
	else
	{
		currentState = State_WanderingAround;
		return;
	}
}

//STATE
//Run-Like-Hell.
//Like the name suggests it, this is an emergency reaction if a collision is imminent.
//We just reverse engines and move back until away from danger. Every escape step
//takes at least 500ms and the robot won't see anything during this period, so in extreme
//cases it may actually hit something when running away from other danger.
//State transition occurs only to Wandring Around if we are clear from danger.
void State_RunLikeHell(TRCBoard1& board)
{
	//Activate beeper to make noise and scare people.
	board.GetMainBoard().powerOn(0);
	
	//Check every sensor to see which direction to go, and move accordingly.
	if( board.GetMainBoard().analog(FL_DISTANCE) > ReactionThreshold_RunLikeHell)
	{
		//uciekaj do tylu
		board.MoveBackward(200);
		board.GetMainBoard().msleep(500);
	}
	else if(board.GetMainBoard().analog(FR_DISTANCE) > ReactionThreshold_RunLikeHell)
	{
		//uciekaj do tylu
		board.MoveBackward(200);
		board.GetMainBoard().msleep(500);
	}
	else if(board.GetMainBoard().analog(BACK_DISTANCE) > ReactionThreshold_RunLikeHell)
	{
		//steruj przed siebie
		board.MoveForward(200);
		board.GetMainBoard().msleep(1500);
	}
	//If we are away from all dangers...
	else
	{
		board.GetMainBoard().powerOff(0);	//Power-off the beeper ;)
		currentState = State_WanderingAround;
	}
}

//STATE
//SignalPresense
//This state is designed to attract attention of people - the robot stops and
//starts flashing LEDs on its board and beeping for 1.25 seconds.
//State transition: to Wandering Around after completion.
void State_SignalPresense(TRCBoard1& board)
{
	board.MoveForward(0);
	for(int i = 0 ; i < 5 ; ++i)
	{
		board.GetMainBoard().led(0, i%2);	//led flashing
		board.GetMainBoard().led(1, (i+1)%2);	//led flashing
		board.GetMainBoard().powerOn(0);	//beeper on
		board.GetMainBoard().msleep(250);	//
		board.GetMainBoard().powerOff(0);	//beeper off
		board.GetMainBoard().msleep(250);	//
	}
	board.GetMainBoard().led(0, 0);
	board.GetMainBoard().led(1, 0);
	
	currentState = State_WanderingAround;
}

//STATE
//LookAround
//Another amusing state, this time the robot beeps and starts rotating in
//random direction for 750-1450 msec.
//State transition: to Wandering Around after completed.
void State_LookAround(TRCBoard1& board)
{
	board.MoveForward(0);
	board.RotateLeft( ((board.BasicRand() % 2)*2 - 1)*75);
	
	board.GetMainBoard().powerOn(0);	//beeper on
	board.GetMainBoard().msleep(250);	//
	board.GetMainBoard().powerOff(0);	//beeper off
	board.GetMainBoard().msleep(500 + board.BasicRand() % 700);	//
	
	currentState = State_WanderingAround;
}

//Program 'entry point':
int main()
{
	//Initialization:
	TRCBoard1 board;	//our board
	
	ReactionThreshold_Evade = board.GetIRThresholdForDistance(ReactionDist_Evade);
	ReactionThreshold_RunLikeHell = board.GetIRThresholdForDistance(ReactionDist_RunLikeHell);
	
	board.GetMainBoard().powerOff(0);	//resets beeper so that it won't be beeping before actual
										//AI is started
	
	board.StartupSequence();	//blocks the program until user presses a button on the board
								//(default: button 0)
	
	currentState = State_WanderingAround;	//default state
	
	//Main loop
	while(1)
	{
		//Just do a state update.
		if(currentState != 0)
		{
			currentState(board);
		}
	}
	
	//End. Program should never get here.
	return 0;
}
